#!/usr/bin/env python

# -*- coding: utf-8 -*-

# Written (W) 2008-2012 Christian Widmer
# Written (W) 2008-2010 Cheng Soon Ong
# Written (W) 2012-2014 Daniel Blanchard, dblanchard@ets.org
# Copyright (C) 2008-2012 Max-Planck-Society, 2012-2014 ETS

# This file is part of GridMap.

# GridMap is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# GridMap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with GridMap.  If not, see <http://www.gnu.org/licenses/>.
"""
Simple web front-end for GridMap

:author: Christian Widmer
:author: Cheng Soon Ong
:author: Dan Blanchard (dblanchard@ets.org)
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import argparse
import logging
import sys
from io import open
from socket import gethostname

import cherrypy

from gridmap.runner import _send_zmq_msg
from gridmap.version import __version__


class WebMonitor(object):
    '''
    A web front-end for the JobMonitor class.
    '''

    @cherrypy.expose
    def index(self):
        return '''
               <form action="list_jobs" method="GET">
               Address of GridMap session: <br />
               <input type="text" name="address" /><br /><br />
               <input type="submit" />
               </form>
               '''.encode()

    @cherrypy.expose
    def list_jobs(self, address):
        """
        display list of jobs
        """
        job_id = -1
        jobs = _send_zmq_msg(job_id, "get_jobs", "", address)

        out_html = '''
            <form action="list_jobs" method="GET">
            Address of GridMap session: <br />
            <input type="text" name="address" /><br /><br />
            <input type="submit" /><br><br><br>
            </form>

            <form action="view_job" method="GET">
            <table border="1">
            <tr><td>sge job id</td><td>job done</td><td>cause of death</td><tr>
            '''

        for job in jobs:
            out_html += ("<tr><td><a href='/view_job?address={0}" +
                         "&job_id={1}'>{1}</td>").format(address, job.id)
            out_html += "<td>{}</td>".format(job.ret is not None)
            out_html += "<td>{}</td>".format(job.cause_of_death)
            out_html += "</tr>"
        out_html += "</table></form>"
        return out_html

    @cherrypy.expose
    def view_job(self, address, job_id):
        """
        display individual job details
        """
        job = _send_zmq_msg(job_id, "get_job", "", address)
        return self.job_to_html(job)

    @staticmethod
    def job_to_html(job):
        """
        display job as html
        """

        # compose error message
        body_text = "job {0}\n<br>".format(job.name)
        body_text += "last timestamp: {0}\n<br>".format(job.timestamp)
        body_text += "num_resubmits: {0}\n<br>".format(job.num_resubmits)
        body_text += "cause_of_death: {0}\n<br>".format(job.cause_of_death)

        if job.heart_beat:
            body_text += ("last memory usage: {0}\n" +
                          "<br>").format(job.heart_beat["memory"])
            body_text += ("last cpu load: {0}\n" +
                          "<br>").format(job.heart_beat["cpu_load"])

        body_text += "host: {}<br><br>\n\n".format(job.host_name)

        if isinstance(job.ret, Exception):
            body_text += "job encountered exception: {0}\n<br>".format(job.ret)
            body_text += "stacktrace: {0}\n<br>\n<br>".format(job.exception)

        # attach log file
        if job.heart_beat:
            with open(job.heart_beat["log_file"], "r") as log_file:
                log_file_attachement = log_file.read().replace("\n", "<br>\n")
            body_text += "<br><br><br>" + log_file_attachement
        return body_text.encode()


def main(argv=None):
    """
    Parse the command line inputs and start web monitor.

    :param argv: List of arguments, as if specified on the command-line.
                 If None, ``sys.argv`` is used instead.
    :type argv: list of str
    """
    # Get command line arguments
    if argv is None:
        argv = sys.argv
    parser = argparse.ArgumentParser(description="Provides a web interface to \
                                                  0MQ job monitor.")
    parser.add_argument('module_dir',
                        help='Directory that contains module containing pickled\
                              function. This will get added to PYTHONPATH \
                              temporarily.', nargs='+')
    parser.add_argument('-p', '--port',
                        help='Port for server to listen on.', type=int,
                        default=8076)
    parser.add_argument('--version', action='version',
                        version='%(prog)s {0}'.format(__version__))
    args = parser.parse_args(argv)

    # Make warnings from built-in warnings module get formatted more nicely
    logging.captureWarnings(True)
    logging.basicConfig(format=('%(asctime)s - %(name)s - %(levelname)s - ' +
                                '%(message)s'))
    logger = logging.getLogger(__name__)

    # Append module directories to path
    for module_dir in args.module_dir:
        logger.info("Appended {0} to PYTHONPATH".format(module_dir))
        sys.path.append(module_dir)

    # Start server
    hostname = gethostname()
    if not isinstance(hostname, bytes):
        hostname = hostname.encode()
    cherrypy.quickstart(WebMonitor(),
                        config={b'global': {b'server.socket_port': args.port,
                                            b'server.socket_host': hostname}})


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

# Written (W) 2008-2012 Christian Widmer
# Written (W) 2008-2010 Cheng Soon Ong
# Written (W) 2012-2014 Daniel Blanchard, dblanchard@ets.org
# Copyright (C) 2008-2012 Max-Planck-Society, 2012-2014 ETS

# This file is part of GridMap.

# GridMap is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# GridMap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with GridMap.  If not, see <http://www.gnu.org/licenses/>.

"""
This module executes pickled jobs on the cluster.

:author: Christian Widmer
:author: Cheng Soon Ong
:author: Dan Blanchard (dblanchard@ets.org)
"""

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import argparse
import logging
import multiprocessing
import os
import random
import socket
import sys
import time
import traceback
from io import open

import psutil
import zmq

from gridmap.conf import HEARTBEAT_FREQUENCY
from gridmap.data import zloads, zdumps


# Set of "not running" job statuses
_SLEEP_STATUSES = {psutil.STATUS_SLEEPING, psutil.STATUS_DEAD,
                   psutil.STATUS_IDLE, psutil.STATUS_STOPPED,
                   psutil.STATUS_ZOMBIE}


# TODO: Refactor this so that there's a class that stores socket, since creating
#       a new one each time is hugely wasteful.
def _send_zmq_msg(job_id, command, data, address):
    """
    simple code to send messages back to host
    (and get a reply back)
    """
    logger = logging.getLogger(__name__)
    context = zmq.Context()
    zsocket = context.socket(zmq.REQ)
    logger.debug('Connecting to JobMonitor (%s)', address)
    zsocket.connect(address)

    host_name = socket.gethostname()
    ip_address = socket.gethostbyname(host_name)

    msg_container = {}
    msg_container["job_id"] = job_id
    msg_container["host_name"] = host_name
    msg_container["ip_address"] = ip_address
    msg_container["command"] = command
    msg_container["data"] = data

    # Send request
    logger.debug('Sending message: %s', msg_container)
    msg_string = zdumps(msg_container)
    zsocket.send(msg_string)

    # Get reply
    msg = zloads(zsocket.recv())

    return msg


def _heart_beat(job_id, address, parent_pid=-1, log_file="", wait_sec=45):
    """
    Infinitely loops and sends information about the currently running job back
    to the ``JobMonitor``.
    """
    while True:
        status = get_job_status(parent_pid, os.getpid())
        if os.path.exists(log_file):
            with open(log_file) as f:
                status["log_file"] = f.read()
        _send_zmq_msg(job_id, "heart_beat", status, address)
        time.sleep(wait_sec)


def get_memory_usage(pid, heart_pid):
    """
    :param pid: Process ID for job whose memory usage we'd like to check.
    :type pid: int
    :param heart_pid: ID of the heartbeat process, which will not be counted
                      toward total.
    :type heart_pid: int

    :returns: Total memory usage of process (and children) in Mb.
    """
    process = psutil.Process(pid)
    mem_total = float(process.memory_info()[0])
    for p in process.children(recursive=True):
        if p.is_running() and p.pid != heart_pid:
            try:
                mem_total += float(p.memory_info()[0])
            except psutil.NoSuchProcess:
                continue
    return mem_total / (1024.0 ** 2.0)


def get_cpu_load(pid, heart_pid):
    """
    :param pid: Process ID for job whose CPU load we'd like to check.
    :type pid: int
    :param heart_pid: ID of the heartbeat process, which will not be counted
                      toward total.
    :type heart_pid: int

    :returns: Tuple of average ratio of CPU time to real time for process and
              its children (excluding heartbeat process), and whether at least
              one of the processes is not sleeping.
    :rtype: (float, bool)
    """
    process = psutil.Process(pid)
    cpu_sum = float(process.cpu_percent())
    running = process.status not in _SLEEP_STATUSES
    num_procs = 1
    for p in process.children(recursive=True):
        # Make sure this process hasn't exited before querying its status
        if p.is_running() and p.pid != heart_pid:
            try:
                cpu_sum += float(p.cpu_percent())
                running = running or (p.status not in _SLEEP_STATUSES)
            except psutil.NoSuchProcess:
                continue
            else:
                num_procs += 1
    return cpu_sum / num_procs, running


def get_job_status(parent_pid, heart_pid):
    """
    Determines the status of the current worker and its machine

    :param parent_pid: Process ID for job whose status we'd like to check.
    :type parent_pid: int
    :param heart_pid: ID of the heartbeat process, which will not be counted
                      toward total.
    :type heart_pid: int

    :returns: Memory and CPU load information for given PID.
    :rtype: dict
    """

    status_container = {}

    if parent_pid != -1:
        status_container["memory"] = get_memory_usage(parent_pid, heart_pid)
        status_container["cpu_load"] = get_cpu_load(parent_pid, heart_pid)

    return status_container


def _run_job(job_id, address):
    """
    Execute the pickled job and produce pickled output.

    :param job_id: Unique ID of job
    :type job_id: int
    :param address: IP address of submitting host.
    :type address: str
    """
    # create heart beat process
    logger = logging.getLogger(__name__)
    parent_pid = os.getpid()
    heart = multiprocessing.Process(target=_heart_beat,
                                    args=(job_id, address, parent_pid,
                                          os.environ['SGE_STDERR_PATH'],
                                          HEARTBEAT_FREQUENCY))
    logger.info("Starting heart beat")
    heart.start()

    wait_sec = random.randint(0, 5)
    logger.info("Waiting %s seconds before fetching input", wait_sec)
    time.sleep(wait_sec)

    try:
        job = _send_zmq_msg(job_id, "fetch_input", None, address)
    except Exception as e:
        # here we will catch errors caused by pickled objects
        # of classes defined in modules not in PYTHONPATH
        logger.error('Could not retrieve input for job {0}'.format(job_id),
                     exc_info=True)

        # send back exception and traceback string
        thank_you_note = _send_zmq_msg(job_id, "store_output",
                                       (e, traceback.format_exc()),
                                       address)
        # stop heartbeat
        heart.terminate()
        return

    logger.debug("Input arguments loaded, starting computation %s", job)

    try:
        # change working directory
        if job.working_dir is not None:
            logger.info("Changing working directory: %s", job.working_dir)
            os.chdir(job.working_dir)

        # run job
        logger.info("Executing job")
        job.execute()
        logger.info("Finished job")

        # send back result
        thank_you_note = _send_zmq_msg(job_id, "store_output", job, address)
        logger.info(thank_you_note)
    # If anything goes wrong, send back the exception
    except Exception as e:
        # send back exception and traceback string
        thank_you_note = _send_zmq_msg(job_id, "store_output",
                                       (e, traceback.format_exc()),
                                       address)
        logger.info(thank_you_note)
    finally:
        # stop heartbeat
        heart.terminate()


def _main():
    """
    Parse the command line inputs and call _run_job
    """

    # Get command line arguments
    parser = argparse.ArgumentParser(description="This wrapper script will run \
                                                  a pickled Python function on \
                                                  some pickled retrieved data \
                                                  via 0MQ. You almost never \
                                                  want to run this yourself.")
    parser.add_argument('home_address',
                        help='IP address of submitting host.')
    parser.add_argument('module_dir',
                        help='Directory that contains module containing pickled\
                              function. This will get added to PYTHONPATH \
                              temporarily.')
    args = parser.parse_args()

    # Make warnings from built-in warnings module get formatted more nicely
    logging.captureWarnings(True)
    logging.basicConfig(format=('%(asctime)s - %(name)s - %(levelname)s - ' +
                                '%(message)s'), level=logging.INFO)
    logger = logging.getLogger(__name__)

    logger.info("Appended {0} to PYTHONPATH".format(args.module_dir))
    sys.path.insert(0, args.module_dir)

    logger.debug("Job ID: %s\tHome address: %s\tModule dir: %s",
                 os.environ['JOB_ID'],
                 args.home_address, args.module_dir)

    # Process the database and get job started
    _run_job(os.environ['JOB_ID'], args.home_address)


if __name__ == "__main__":
    _main()



# -*- coding: utf-8 -*-

# Written (W) 2008-2012 Christian Widmer
# Written (W) 2008-2010 Cheng Soon Ong
# Written (W) 2012-2014 Daniel Blanchard, dblanchard@ets.org
# Copyright (C) 2008-2012 Max-Planck-Society, 2012-2014 ETS

# This file is part of GridMap.

# GridMap is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# GridMap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with GridMap.  If not, see <http://www.gnu.org/licenses/>.
"""
Global settings for GridMap. All of these settings can be overridden by
specifying environment variables with the same name.

:author: Christian Widmer
:author: Cheng Soon Ong
:author: Dan Blanchard (dblanchard@ets.org)

:var USE_MEM_FREE: Does your cluster support specifying how much memory a job
                   will use via mem_free? (Default: ``False``)
:var DEFAULT_QUEUE: The default job scheduling queue to use.
                    (Default: ``all.q``)
:var CREATE_PLOTS: Should we plot cpu and mem usage and send via email?
                   (Default: ``True``)
:var SEND_ERROR_MAIL: Should we send error emails?
                      (Default: ``True``)
:var SMTP_SERVER: SMTP server for sending error emails.
                  (Default: last three sections of the current machine's fully
                  qualified domain name)
:var ERROR_MAIL_SENDER: Sender address to use for error emails.
                        (Default: ``error@gridmap.py``)
:var ERROR_MAIL_RECIPIENT: Recipient address for error emails.
                           (Default: $USER@$HOST, where $USER is the current
                           user's username, and $HOST is the last two sections
                           of the current machine's fully qualified domain name,
                           or just the hostname if it does not contain periods.)
:var MAX_TIME_BETWEEN_HEARTBEATS: How long should we wait (in seconds) for a
                                  heartbeat before we consider a job dead?
                                  (Default: 90)
:var IDLE_THRESHOLD: Percent CPU utilization (ratio of CPU time to real time
                     * 100) that a process must drop below to be considered not
                     running.
                     (Default: 1.0)
:var MAX_IDLE_HEARTBEATS: Number of heartbeats we can receive where the process
                          has <= IDLE_THRESHOLD CPU utilization and is sleeping
                          before we consider the process dead. (Default: 3)
:var NUM_RESUBMITS: How many times can a particular job can die, before we give
                    up. (Default: 3)
:var CHECK_FREQUENCY: How many seconds pass before we check on the status of a
                      particular job in seconds. (Default: 15)
:var HEARTBEAT_FREQUENCY: How many seconds pass before jobs on the cluster send
                          back heart beats to the submission host.
                          (Default: 10)
:var DEFAULT_PAR_ENV: Default parallel environment to use 
                      (Default: ``smp``)
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import logging
import os
from socket import gethostname


# Check if certain libraries are present
try:
    import drmaa
    DRMAA_PRESENT = True
except (ImportError, OSError, RuntimeError):
    logger = logging.getLogger(__name__)
    logger.warning('Could not import drmaa. Only local multiprocessing ' +
                   'supported.')
    DRMAA_PRESENT = False

# plot cpu and mem usage and send via email
CREATE_PLOTS = 'TRUE' == os.getenv('CREATE_PLOTS', 'True').upper()
if CREATE_PLOTS:
    try:
        import matplotlib
    except ImportError:
        logger = logging.getLogger(__name__)
        logger.warning('Could not import matplotlib. No plots will be created' +
                       ' in debug emails.')
        CREATE_PLOTS = False

# Global settings ####
# email settings
SEND_ERROR_MAIL = 'TRUE' == os.getenv('SEND_ERROR_MAIL', 'True').upper()
SMTP_SERVER = os.getenv('SMTP_SERVER', '.'.join(gethostname().split('.')[-3:]))
ERROR_MAIL_SENDER = os.getenv('ERROR_MAIL_SENDER', 'error@gridmap.py')
ERROR_MAIL_RECIPIENT = os.getenv('ERROR_MAIL_RECIPIENT',
                                 '{}@{}'.format(os.getenv('USER'),
                                                '.'.join(gethostname().split('.')[-2:])))

# how much time can pass between heartbeats, before
# job is assummed to be dead in seconds
MAX_TIME_BETWEEN_HEARTBEATS = int(os.getenv('MAX_TIME_BETWEEN_HEARTBEATS',
                                            '90'))

IDLE_THRESHOLD = float(os.getenv('IDLE_THRESHOLD', '1.0'))
MAX_IDLE_HEARTBEATS = int(os.getenv('MAX_IDLE_HEARTBEATS', '3'))

# defines how many times can a particular job can die,
# before we give up
NUM_RESUBMITS = int(os.getenv('NUM_RESUBMITS', '3'))

# check interval: how many seconds pass before we check
# on the status of a particular job in seconds
CHECK_FREQUENCY = int(os.getenv('CHECK_FREQUENCY', '15'))

# heartbeat frequency: how many seconds pass before jobs
# on the cluster send back heart beats to the submission
# host
HEARTBEAT_FREQUENCY = int(os.getenv('HEARTBEAT_FREQUENCY', '10'))

# Is mem_free configured properly on the cluster?
USE_MEM_FREE = 'TRUE' == os.getenv('USE_MEM_FREE', 'False').upper()

# Which queue should we use by default
DEFAULT_QUEUE = os.getenv('DEFAULT_QUEUE', 'all.q')

# Where shall we use as temporary_directory
DEFAULT_TEMP_DIR = os.getenv('DEFAULT_TEMP_DIR', '/scratch/')

# Which parallel environment to use by default
DEFAULT_PAR_ENV = os.getenv('DEFAULT_PAR_ENV', "smp")
